/**
 *  @file      Transactional.java
 *  @author    Stefan Reichert
 *
 *  Copyright (c)
 *  BTC Embedded Systems AG
 *  Oldenburg, Germany
 *
 *  --------------------------------
 *  $$Author: ${user} $$
 *  $$Date: ${date} ${time} $$
 *  $$Rev: revision number $$
 *  --------------------------------
 */
package com.btc.ep.base.transactions.internal;

import java.util.UUID;

import javax.persistence.EntityManager;

// ---------------------------------------------------------------------------
// Implementation
// ---------------------------------------------------------------------------

/**
 * This class represents the container for transactional object that is attached to the thread.
 */
public class TransactionData {

    /** The entity manger to be used for database access. */
    private final EntityManager entityManager;
    /** Marker for rollback-only transactions. */
    private final boolean isRollbackOnly;
    /** This transaction's uuid */
    private final String transactionId = UUID.randomUUID().toString();
    /** The current depth of nested transactions. */
    private int nestingDepth = 0;

    /**
     * Constructor for {@link TransactionData} having the entity manager as parameter.
     *
     * @param entityManager the entity manager instance to be used for a transaction
     * @param isRollbackOnly marks the transaction as rollback-only, i.e. all data will be rolled back on commit
     */
    public TransactionData(EntityManager entityManager, boolean isRollbackOnly) {
        this.entityManager = entityManager;
        this.isRollbackOnly = isRollbackOnly;
    }

    /**
     * Brief Function description
     *
     * @return
     */
    public EntityManager getEntityManager() {
        return entityManager;
    }

    /**
     * Returns whether the current call stack is in <i>nested</i> state. That is true if there are more that one methods
     * on the call stack, that represent a unit of work (i.e. the nesting depth is greater than zero).
     *
     * @return whether the current call stack is in a <i>nested state</i>
     */
    public boolean isInNestedTransactionState() {
        return nestingDepth > 0;
    }

    /**
     * Increments the current nesting depth by one.
     *
     * @return the new nesting depth
     */
    public int incrementNestingDepthByOne() {
        nestingDepth = nestingDepth + 1;
        return nestingDepth;
    }

    /**
     * Decrements the current nesting depth by one.
     *
     * @return the new nesting depth
     */
    public int decrementNestingDepthByOne() {
        nestingDepth = nestingDepth - 1;
        return nestingDepth;
    }

    /**
     * Returns the transaction's unique identifier.
     *
     * @return the transaction's identifier
     */
    public String getTransactionId() {
        return transactionId;

    }

    /**
     * Returns whether this transaction is marked as rollback-only.
     *
     * @return true if transaction is marked as rollback-only, false otherwise
     */
    public boolean isRollbackOnly() {
        return isRollbackOnly;

    }

}

// ---------------------------------------------------------------------------
// END OF FILE
// ---------------------------------------------------------------------------
