/**
 *  @file      TransactionManagerImpl.java
 *  @author    Stefan Reichert
 *
 *  Copyright (c)
 *  BTC Embedded Systems AG
 *  Oldenburg, Germany
 *
 *  --------------------------------
 *  $$Author: ${user} $$
 *  $$Date: ${date} ${time} $$
 *  $$Rev: revision number $$
 *  --------------------------------
 */
package com.btc.ep.base.transactions.internal;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.EntityTransaction;
import javax.persistence.FlushModeType;

import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Reference;
import org.osgi.service.component.annotations.ReferenceCardinality;

import com.btc.ep.base.transactions.EntityManagerFactoryProvider;
import com.btc.ep.base.transactions.TransactionContext;
import com.btc.ep.base.transactions.TransactionException;
import com.btc.ep.base.transactions.TransactionManager;

//---------------------------------------------------------------------------
//Implementation
//---------------------------------------------------------------------------

/**
 * This class represents the transaction manager implementation. It provides a TransactionContext based on a
 * {@link ThreadLocal}.<br>
 * The implementation implements both interfaces {@link TransactionManager} and {@link TransactionContext}. This is for
 * convenience reasons.<br>
 * The implementation is registered at the OSGi service bus with both interfaces, so it will be used by both aspect and
 * other components depending on transactional objects such as the {@link EntityManager}.
 */
@Component (name = "transactionmanager", service = {TransactionManager.class, TransactionContext.class })
public class TransactionManagerImpl implements TransactionManager, TransactionContext {

    /** The logger instance to be used. */

    /** The thread local field holding the respective thread's {@link TransactionData}. */
    private ThreadLocal<TransactionData> threadLocal = new ThreadLocal<TransactionData>();

    /**
     * The {@link EntityManagerFactoryProvider} that provides access to the {@link EntityManagerFactory} used for
     * creating an {@link EntityManager} per transaction.
     */
    private EntityManagerFactoryProvider entityManagerFactoryProvider;

    /**
     * Returns the {@link EntityManagerFactory} obtained from the {@link EntityManagerFactoryProvider}.
     *
     * @return an {@link EntityManagerFactory} instance
     * 
     * @throws IllegalStateException in case no {@link EntityManagerFactoryProvider} is available or the
     *             {@link EntityManagerFactoryProvider} does not provide an {@link EntityManagerFactory}
     */
    private EntityManagerFactory getEntityManagerFactory() {
        if (entityManagerFactoryProvider == null) {
            throw new IllegalStateException(
                "transaction handling failed - no entity manager factory provider available");
        }
        EntityManagerFactory entityManagerFactory = entityManagerFactoryProvider.getEntityManagerFactory();
        if (entityManagerFactory == null) {
            throw new IllegalStateException("no entity manager factory available");
        }
        return entityManagerFactory;
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.btc.ep.base.transactions.TransactionContext#getEntityManager()
     */
    @Override
    public EntityManager getEntityManager() throws TransactionException {
        return getTransactionData().getEntityManager();
    }

    /**
     * Returns the current thread's transaction data object.
     *
     * @return the current thread's transaction data object
     * @throws IllegalStateException
     */
    private TransactionData getTransactionData() {
        TransactionData transactionData = threadLocal.get();
        if (transactionData == null) {
            throw new TransactionException("no active transaction");
        }
        return transactionData;
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.btc.ep.base.transactions.internal.TransactionManager#init(org.aspectj.lang.Signature)
     */
    @Override
    public TransactionContext init(String jointPointId, boolean isRollbackOnly) throws IllegalStateException {
        TransactionData transactionData = threadLocal.get();
        if (transactionData == null) {
            doInit(jointPointId, isRollbackOnly);
        } else {
            transactionData.incrementNestingDepthByOne();
            if (isRollbackOnly && !transactionData.isRollbackOnly()) {
            }
        }
        return this;
    }

    private void doInit(String jointPointId, boolean isRollbackOnly) {
        EntityManager entityManager = getEntityManagerFactory().createEntityManager();
        entityManager.setFlushMode(FlushModeType.AUTO);
        EntityTransaction transaction = entityManager.getTransaction();
        transaction.begin();
        TransactionData transactionData = new TransactionData(entityManager, isRollbackOnly);
        threadLocal.set(transactionData);
    }

    private String getLogPrefix(String jointPointId, TransactionData transactionData) {
        return transactionData.getTransactionId() + " # " + jointPointId;
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.btc.ep.base.transactions.internal.TransactionManager#commit(org.aspectj.lang.Signature)
     */
    @Override
    public void commit(String jointPointId) {
        TransactionData transactionData = getTransactionData();
        if (!transactionData.isInNestedTransactionState()) {
            EntityTransaction transaction = getEntityManager().getTransaction();
            if (transaction.isActive()) {
                doCommit(jointPointId, transactionData);
            } else {
            }
        } else {
        }
    }

    private void doCommit(String jointPointId, TransactionData transactionData) {
        EntityTransaction transaction = getEntityManager().getTransaction();
        if (transactionData.isRollbackOnly()) {
            transaction.rollback();
        } else {
            try {
                transaction.commit();
            } catch (RuntimeException runtimeException) {
                throw new TransactionException(String.format("committing transaction %s failed",
                    transactionData.getTransactionId()), runtimeException);
            }
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.btc.ep.base.transactions.internal.TransactionManager#rollback(org.aspectj.lang.Signature,
     *      java.lang.RuntimeException)
     */
    @Override
    public void rollback(String jointPointId, RuntimeException cause) {
        TransactionData transactionData = getTransactionData();
        if (!transactionData.isInNestedTransactionState()) {
            EntityTransaction transaction = getEntityManager().getTransaction();
            if (transaction.isActive()) {
                doRollback(jointPointId, cause, transactionData);
            } else {
            }
        } else {
        }
    }

    private void doRollback(String jointPointId, RuntimeException cause, TransactionData transactionData) {
        try {
            transactionData.getEntityManager().getTransaction().rollback();
        } catch (RuntimeException runtimeException) {
            throw new TransactionException(String.format("rolling back transaction %s failed",
                transactionData.getTransactionId()), runtimeException);
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.btc.ep.base.transactions.internal.TransactionManager#close(org.aspectj.lang.Signature)
     */
    @Override
    public void close(String jointPointId) {
        TransactionData transactionData = getTransactionData();
        if (!transactionData.isInNestedTransactionState()) {
            doClose(jointPointId, transactionData);
        } else {
            transactionData.decrementNestingDepthByOne();
        }
    }

    private void doClose(String jointPointId, TransactionData transactionData) {
        try {
            transactionData.getEntityManager().close();
        } catch (RuntimeException runtimeException) {
            throw new TransactionException(String.format("closing transaction %s failed",
                transactionData.getTransactionId()), runtimeException);
        } finally {
            threadLocal.remove();
        }
    }

    /**
     * Sets the required {@link EntityManagerFactoryProvider}. This method is intended to be used by the DI framework
     * only.
     *
     * @param entityManagerFactoryProvider The required {@link EntityManagerFactoryProvider} collaborator
     */
    @Reference (cardinality = ReferenceCardinality.OPTIONAL)
    public void setEntityManagerFactoryProvider(EntityManagerFactoryProvider entityManagerFactoryProvider) {
        this.entityManagerFactoryProvider = entityManagerFactoryProvider;
    }
}
