/**
 *  Copyright (c) 2014, 
 *  BTC Embedded Systems AG, Oldenburg, Germany  
 *  All rights reserved
 *  --------------------------------------------------------- 
 *  $Author: $
 *  $Date:  $
 *  $Revision: $
 **/

package com.btc.ep.dal.provider.impl;

import java.io.File;
import java.io.IOException;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.StandardCopyOption;
import java.nio.file.attribute.BasicFileAttributes;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;
import javax.persistence.TypedQuery;


import com.btc.ep.base.transactions.EntityManagerFactoryProvider;
import com.btc.ep.dal.provider.PersistenceManagerProvider;

/**
 * Provider for objectDB.
 * 
 * @author jaceck
 * @author Simon Kollmann
 * 
 */
public class OdbPersistenceManagerProvider implements PersistenceManagerProvider, EntityManagerFactoryProvider {


    public static final String ID = "objectDB";
    private static final String PERSISTENCE_UNIT_ODB = "eppdb";
    private static final String DB_FILE = "eppdb.odb";

    private EntityManagerFactory entityManagerFactory;

    
    @Override
    public String getId() {
        return ID;
    }

    @Override
    public EntityManagerFactory getEntityManagerFactory() {
        if (entityManagerFactory == null) {

            //
            String preConfiguredODBConfigFile = System.getProperty("objectdb.conf");
            if (preConfiguredODBConfigFile != null) {
            } 

            //check and, if required, set objectdb.home
            String objectdbHome = System.getProperty("objectdb.home");
            if (objectdbHome != null) {
            } else {
                // set ObjectDB Home
                File dbPath = new File(DB_FILE).getAbsoluteFile().getParentFile();

                try {
                    objectdbHome = dbPath.getCanonicalPath();
                } catch (IOException e) {
                }

                if (objectdbHome != null) {
                    System.setProperty("objectdb.home", objectdbHome);
                } else {
                }
            }
            
            this.entityManagerFactory = Persistence.createEntityManagerFactory(PERSISTENCE_UNIT_ODB);

        }
        return this.entityManagerFactory;
    }

    @Override
    public File getDBPath() {
        File dbPath = null;
        if (entityManagerFactory != null) {
            String path = (String)entityManagerFactory.getProperties().get("objectdb.connection.path");
            dbPath = new File(path);
        } else {
            String path = System.getProperty("objectdb.home");
            if (path != null) {
                dbPath = new File(path, DB_FILE);
            }
        }
        return dbPath;
    }

    @Override
    public void backup(final File backupFile) throws IOException {
        
        final File dbFile = getDBPath();

        if (this.entityManagerFactory != null) {
            EntityManager em = entityManagerFactory.createEntityManager();
            Path backupDir;
            try {
                TypedQuery<Thread> backupQuery = em.createQuery("objectdb backup", Thread.class);
    
                backupDir = Files.createTempDirectory(dbFile.getParentFile().toPath(), "BAK");
                backupQuery.setParameter("target", backupDir.toFile());
                Thread bt = (Thread)backupQuery.getSingleResult();
                try {
                    bt.join();
                } catch (InterruptedException e) {
                    throw new RuntimeException(e);
                }
            } finally {
                em.close();
            }
            
            // TODO is it possible to omit this second copy? 
            // the first copy above could be copy directly to the desirable directory
            // it is possible advantage for big database files
            Files.walkFileTree(backupDir, new SimpleFileVisitor<Path>() {

                @Override
                public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
                    if (file.getFileName().toString().equals(dbFile.getName())) {
                        Files.copy(file, backupFile.toPath(), StandardCopyOption.REPLACE_EXISTING);
                    }
                    Files.delete(file);
                    return FileVisitResult.CONTINUE;
                }

                @Override
                public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
                    Files.delete(dir);
                    return FileVisitResult.CONTINUE;
                }
            });

        } else {
            Files.copy(dbFile.toPath(), backupFile.toPath(), StandardCopyOption.REPLACE_EXISTING);
        }
    }

    @Override
    public void close() {
        if (entityManagerFactory == null) {
            return;
        }
        this.entityManagerFactory.close();
        this.entityManagerFactory = null;
    }

    @Override
    public void clearFiles() {
        File path = getDBPath();
        close();
        if (path != null) {
            File path2 = new File(path.getPath() + "$");
            try {
                Files.deleteIfExists(path.toPath());
                Files.deleteIfExists(path2.toPath());
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
    }

    @Override
    public void openFile(File path) throws Exception {
        File dbpath = getDBPath();

        boolean opened = entityManagerFactory != null;
        
        clearFiles();
        
        Files.copy(path.toPath(), dbpath.toPath());
        
        if (opened) {
            getEntityManagerFactory();
        }
    }
}
